#include <iostream>
#include <array>
#include <vector>
#include <set>
#include <algorithm>
#include <numeric>
#include <ext/pb_ds/assoc_container.hpp>
#include <ext/pb_ds/tree_policy.hpp>

using namespace std;
using namespace __gnu_pbds;

using ordered_set = tree<int, null_type, less<int>, rb_tree_tag, tree_order_statistics_node_update>;

struct DSU
{
    vector<int> parent, rank;
    vector<ordered_set> vertices;
    
    DSU(int n)
    {
        parent.resize(n + 1);
        rank.resize(n + 1);
        vertices.resize(n + 1);
        iota(parent.begin(), parent.end(), 0);
        for (int i = 1; i <= n; ++i) vertices[i].insert(i);
    }
    
    int getParent(int v)
    {
        while (v != parent[v]) v = parent[v];
        return v;
    }
    
    void unite(int a, int b)
    {
        a = getParent(a);
        b = getParent(b);
        if (a == b) return;
        if (rank[a] < rank[b]) swap(a, b);
        if (rank[a] == rank[b]) ++rank[a];
        parent[b] = a;
        if (vertices[a].size() < vertices[b].size()) vertices[a].swap(vertices[b]);
        for (int i : vertices[b]) vertices[a].insert(i);
    }

    int get(int a, int k)
    {
        a = getParent(a);
        if (k > vertices[a].size()) return -1;
        return *vertices[a].find_by_order(k - 1);
    }
    
};

int main()
{
    cin.tie(0)->sync_with_stdio(0);
    int n, q, g;
    cin >> n >> q >> g;
    DSU dsu(n);
    while (q--)
    {
        int t;
        cin >> t;
        if (t == 1)
        {
            int v, k;
            cin >> v >> k;
            cout << dsu.get(v, k) << '\n';
        }
        else if (t == 2)
        {
            int v, u;
            cin >> v >> u;
            dsu.unite(v, u);
        }
    }
    return 0;
}
