#include <iostream>
#include <array>
#include <vector>
#include <set>
#include <algorithm>
#include <numeric>
#include <ext/pb_ds/assoc_container.hpp>
#include <ext/pb_ds/tree_policy.hpp>
#include <stack>
#include <cmath>

using namespace std;
using namespace __gnu_pbds;

using ordered_set = tree<int, null_type, less<int>, rb_tree_tag, tree_order_statistics_node_update>;

struct DSU
{
    struct op
    {
        int u, v;
        int rank, parent;
    };
    
    vector<int> parent, rank;
    vector<ordered_set> vertices;
    stack<op> ops;
    
    DSU(int n = -1)
    {
        parent.resize(n + 1);
        rank.resize(n + 1);
        vertices.resize(n + 1);
        iota(parent.begin(), parent.end(), 0);
        for (int i = 1; i <= n; ++i) vertices[i].insert(i);
    }
    
    int getParent(int v)
    {
        while (v != parent[v]) v = parent[v];
        return v;
    }
    
    void unite(int a, int b)
    {
        a = getParent(a);
        b = getParent(b);
        if (a == b)
        {
            op r;
            r.u = -1;
            ops.push(r);
            return;
        }
        if (rank[a] < rank[b]) swap(a, b);
        op r;
        r.u = a;
        r.v = b;
        r.rank = rank[a];
        r.parent = parent[b];
        ops.push(r);
        if (rank[a] == rank[b]) ++rank[a];
        parent[b] = a;
        //if (vertices[a].size() < vertices[b].size()) vertices[a].swap(vertices[b]);
        for (int i : vertices[b]) vertices[a].insert(i);
    }
    
    int get(int a, int k)
    {
        a = getParent(a);
        if (k > vertices[a].size()) return -1;
        return *vertices[a].find_by_order(k - 1);
    }
    
    void rollBack()
    {
        op r = ops.top();
        ops.pop();
        if (r.u == -1) return;
        for (int i : vertices[r.v]) vertices[r.u].erase(i);
        rank[r.u] = r.rank;
        parent[r.v] = r.parent;
    }
    
};

int main()
{
    cin.tie(0)->sync_with_stdio(0);
    int n, q, g;
    cin >> n >> q >> g;
    DSU dsu(n);
    int every = sqrtl(q) + 1;
    vector<DSU> saved(q / every + 5);
    for (int i = 1; i <= q; ++i)
    {
        if ((i - 1) % every == 0) saved[i / every] = dsu;
        int t;
        cin >> t;
        if (t == 1)
        {
            int v, k;
            cin >> v >> k;
            cout << dsu.get(v, k) << '\n';
        }
        else if (t == 2)
        {
            int v, u;
            cin >> v >> u;
            dsu.unite(v, u);
        }
        else
        {
            int x;
            cin >> x;
            if (x / every == i / every)
            {
                for (int j = 0; j < i - x; ++j) dsu.rollBack();
            }
            else
            {
                int val = (x + every - 1) / every;
                dsu = saved[val];
                for (int j = 0; j < val * every - x; ++j) dsu.rollBack();
            }
        }
    }
    return 0;
}
